from decimal import Decimal
from dataclasses import dataclass

from aiomexc.enums import OrderStatus, OrderType, OrderSide


@dataclass
class Order:
    symbol: str
    order_id: str
    order_list_id: int | None
    client_order_id: str | None
    price: Decimal
    orig_qty: Decimal
    executed_qty: Decimal
    cummulative_quote_qty: Decimal
    status: OrderStatus
    time_in_force: int | None
    type: OrderType
    side: OrderSide
    stop_price: Decimal | None
    time: int
    update_time: int | None
    is_working: bool
    orig_quote_order_qty: Decimal | None



from typing import Any

from aiomexc.methods import MexcMethod
from aiomexc.types import MexcType


class MexcClientError(Exception):
    """
    Base exception for all mexc client errors.
    """


class DetailedMexcClientError(MexcClientError):
    """
    Base exception for all mexc client errors with a detailed message.
    """

    def __init__(self, message: str):
        self.message = message

    def __str__(self) -> str:
        return self.message

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.message})"


class MexcAPIError(DetailedMexcClientError):
    """
    Base exception for all mexc API errors.
    """

    label: str = "Mexc server says"

    def __init__(
        self,
        method: MexcMethod[MexcType],
        message: str,
    ) -> None:
        super().__init__(message=message)
        self.method = method

    def __str__(self) -> str:
        original_message = super().__str__()
        return f"{self.label} - {original_message}"


class MexcBadRequest(MexcAPIError):
    """
    Exception raised when request is malformed.
    """


class MexcNotFound(MexcAPIError):
    """
    Exception raised when order not found.
    """


class MexcApiKeyInvalid(MexcAPIError):
    """
    Exception raised when API key is invalid.
    """


class MexcApiKeyMissing(MexcAPIError):
    """
    Exception raised when API key is missing.
    """


class MexcApiInvalidListenKey(MexcAPIError):
    """
    Exception raised when listen key is invalid.
    """


class ClientDecodeError(MexcClientError):
    """
    Exception raised when client can't decode response. (Malformed response, etc.)
    """

    def __init__(self, message: str, original: Exception, data: Any) -> None:
        self.message = message
        self.original = original
        self.data = data

    def __str__(self) -> str:
        original_type = type(self.original)
        return (
            f"{self.message}\n"
            f"Caused from error: "
            f"{original_type.__module__}.{original_type.__name__}: {self.original}\n"
            f"Content: {self.data}"
        )


class MexcWsStreamsLimit(MexcClientError):
    """
    Exception raised when too many streams are subscribed.
    """


class MexcWsNoStreamsProvided(MexcClientError):
    """
    Exception raised when no streams are provided.
    """


class MexcWsNoCredentialsProvided(MexcClientError):
    """
    Exception raised when no credentials are provided.
    """


class MexcWsInvalidStream(MexcClientError):
    """
    Exception raised when an invalid stream is provided.
    """

    def __init__(self, stream: str):
        self.stream = stream

    def __str__(self) -> str:
        return f"Invalid stream: {self.stream}"


class MexcWsPrivateStream(MexcClientError):
    """
    Exception raised when a private stream is provided to a public connection.
    """

    def __init__(self, stream: str):
        self.stream = stream

    def __str__(self) -> str:
        return f"Private stream: {self.stream}"


class MexcWsConnectionClosed(MexcClientError):
    """
    Exception raised when the connection is closed.
    """

    def __str__(self) -> str:
        return "Connection closed"


class MexcWsConnectionError(MexcClientError):
    """
    Exception raised when the connection is not established.
    """

    def __str__(self) -> str:
        return "Connection error"


class MexcWsConnectionNotEstablished(MexcClientError):
    """
    Exception raised when the connection is not established.
    """

    def __str__(self) -> str:
        return "Connection not established"



import json
import time
import hmac
import hashlib

from urllib.parse import urlencode
from abc import ABC, abstractmethod
from typing import Callable, Any, Final, cast
from dataclasses import dataclass

from http import HTTPStatus
from aiomexc.methods import MexcMethod
from aiomexc.types import MexcResult, MexcType
from aiomexc.exceptions import (
    MexcBadRequest,
    MexcNotFound,
    MexcApiKeyInvalid,
    MexcApiKeyMissing,
    MexcAPIError,
    MexcApiInvalidListenKey,
)
from aiomexc.retort import _retort

_JsonLoads = Callable[..., Any]
_JsonDumps = Callable[..., str]
DEFAULT_TIMEOUT: Final[float] = 60.0


@dataclass
class Credentials:
    access_key: str
    secret_key: str


class BaseSession(ABC):
    def __init__(
        self,
        json_loads: _JsonLoads = json.loads,
        json_dumps: _JsonDumps = json.dumps,
        timeout: float = DEFAULT_TIMEOUT,
        recv_window: int = 30000,
    ):
        self.json_loads = json_loads
        self.json_dumps = json_dumps
        self.timeout = timeout
        self.recv_window = recv_window

        self._map_error_code_to_exception = {
            700004: MexcBadRequest,  # Param 'origClientOrderId' or 'orderId' must be sent, but both were empty/null
            400: MexcBadRequest,  # api key required
            -1121: MexcBadRequest,  # Invalid symbol
            -2013: MexcNotFound,  # Order does not exist
            10072: MexcApiKeyInvalid,  # Invalid API-Key format
            402: MexcApiKeyMissing,  # API-Key missing, not sure
            730708: MexcApiInvalidListenKey,  # Invalid listen key
        }

    def encrypt_params(self, secret_key: str, params: dict | None = None) -> dict:
        """
        Encrypt params for auth required requests

        :param secret_key: Secret key
        :param params: Params to encrypt
        :return: Encrypted params
        """
        if params is not None:
            params = {k: v for k, v in params.items() if v is not None}
        else:
            params = {}

        params["recvWindow"] = self.recv_window
        params["timestamp"] = int(time.time() * 1000)  # timestamp in milliseconds

        encoded_params = urlencode(params)
        params["signature"] = hmac.new(
            secret_key.encode(), encoded_params.encode(), hashlib.sha256
        ).hexdigest()
        return params

    def check_response(
        self, method: MexcMethod[MexcType], status_code: int, json_data: dict
    ) -> MexcResult[MexcType]:
        response_type = MexcResult[method.__returning__]
        response = _retort.load(json_data, response_type)

        if HTTPStatus.OK <= status_code <= HTTPStatus.IM_USED and response.ok:
            return response

        message = cast(str, response.msg)

        if exception_cls := self._map_error_code_to_exception.get(status_code):
            raise exception_cls(
                method=method,
                message=message,
            )

        raise MexcAPIError(
            method=method,
            message=message,
        )

    @abstractmethod
    async def make_signed_request(
        self,
        method: MexcMethod[MexcType],
        credentials: Credentials,
        timeout: float | None = None,
    ) -> MexcType:  # pragma: no cover
        """
        Make signed request to Mexc API

        :param method: Method instance
        :param credentials: Credentials instance
        :param timeout: Request timeout
        :return:
        :raise MexcAPIError:
        """
        pass

    @abstractmethod
    async def make_request(
        self,
        method: MexcMethod[MexcType],
        timeout: float | None = None,
    ) -> MexcType:  # pragma: no cover
        """
        Make request to Mexc API

        :param method: Method instance
        :param timeout: Request timeout
        :param credentials: Optional credentials to use for this specific request
        :return:
        :raise MexcAPIError:
        """
        pass

    @abstractmethod
    async def close(self) -> None:  # pragma: no cover
        """
        Close client session
        """
        pass



import asyncio
import ssl

from typing import Any, Dict, Type, cast
from urllib.parse import urljoin
import certifi

from aiohttp import ClientSession, TCPConnector, ClientTimeout
from aiohttp.hdrs import USER_AGENT
from aiohttp.http import SERVER_SOFTWARE

from aiomexc.methods import MexcMethod
from aiomexc.types import MexcType
from aiomexc.__meta__ import __version__
from aiomexc import loggers
from aiomexc.retort import _retort

from .base import BaseSession, Credentials


class AiohttpSession(BaseSession):
    def __init__(self, limit: int = 100, **kwargs: Any):
        super().__init__(**kwargs)

        self._session: ClientSession | None = None
        self._connector_type: Type[TCPConnector] = TCPConnector
        self._connector_init: Dict[str, Any] = {
            "ssl": ssl.create_default_context(cafile=certifi.where()),
            "limit": limit,
            "ttl_dns_cache": 3600,  # Workaround for https://github.com/aiogram/aiogram/issues/1500
        }
        self._should_reset_connector = True  # flag determines connector state
        self._base_url = "https://api.mexc.com/api/v3/"
        self._headers = {
            "Content-Type": "application/json",
        }

    async def create_session(self) -> ClientSession:
        if self._should_reset_connector:
            await self.close()

        if self._session is None or self._session.closed:
            self._session = ClientSession(
                connector=self._connector_type(**self._connector_init),
                headers={
                    USER_AGENT: f"{SERVER_SOFTWARE} mexc-api/{__version__}",
                },
            )
            self._should_reset_connector = False

        return self._session

    async def close(self) -> None:
        if self._session is not None and not self._session.closed:
            await self._session.close()

            # Wait 250 ms for the underlying SSL connections to close
            # https://docs.aiohttp.org/en/stable/client_advanced.html#graceful-shutdown
            await asyncio.sleep(0.25)

    async def _request(
        self,
        method: MexcMethod[MexcType],
        params: dict[str, Any],
        headers: dict[str, Any],
        timeout: float | None = None,
    ) -> MexcType:
        session = await self.create_session()
        url = urljoin(self._base_url, method.__api_method__)

        loggers.client.debug(
            "Requesting %s %s with params %s", method.__api_http_method__, url, params
        )

        try:
            async with session.request(
                method.__api_http_method__,
                url,
                params=params,
                headers=headers,
                timeout=ClientTimeout(
                    total=self.timeout if timeout is None else timeout
                ),
            ) as resp:
                raw_result = await resp.json()

                if isinstance(
                    raw_result, dict
                ):  # we can trust the api that the error will not be returned in the list
                    api_code = int(raw_result.get("code", 200))
                    msg = raw_result.get("msg")
                else:
                    api_code = 200
                    msg = None

                wrapped_result = {
                    "ok": resp.ok,
                    "msg": msg,
                    "code": api_code,
                    "result": raw_result if api_code == 200 else None,
                }  # this is needed, because mexc api don't have stable response structure
        except asyncio.TimeoutError:
            raise

        loggers.client.debug("Response: %s", wrapped_result)
        response = self.check_response(method, api_code, wrapped_result)
        return cast(MexcType, response.result)

    async def make_signed_request(
        self,
        method: MexcMethod[MexcType],
        credentials: Credentials,
        timeout: float | None = None,
    ) -> MexcType:
        params = _retort.dump(method)
        headers = self._headers.copy()

        if method.__requires_auth__:
            params = self.encrypt_params(credentials.secret_key, params)
            headers["X-MEXC-APIKEY"] = credentials.access_key

        return await self._request(method, params, headers, timeout)

    async def make_request(
        self,
        method: MexcMethod[MexcType],
        timeout: float | None = None,
    ) -> MexcType:
        return await self._request(method, _retort.dump(method), self._headers, timeout)

    async def __aenter__(self) -> "AiohttpSession":
        await self.create_session()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        await self.close()



from adaptix import NameStyle, Retort, name_mapping

from aiomexc.methods import (
    MexcMethod,
    QueryOrder,
    CreateListenKey,
    GetListenKeys,
    ExtendListenKey,
    DeleteListenKey,
)
from aiomexc.types import Order, AccountInformation, TickerPrice, ListenKey, ListenKeys

type_recipes = [
    name_mapping(
        mexc_type,
        name_style=NameStyle.CAMEL,
    )
    for mexc_type in [
        Order,
        AccountInformation,
        TickerPrice,
        ListenKey,
        ListenKeys,
    ]
]

_retort = Retort(
    recipe=[
        name_mapping(
            MexcMethod,
            name_style=NameStyle.CAMEL,
            omit_default=True,
        ),
        name_mapping(
            QueryOrder,
            name_style=NameStyle.CAMEL,
            omit_default=True,
        ),
        name_mapping(
            CreateListenKey,
            name_style=NameStyle.CAMEL,
            omit_default=True,
        ),
        name_mapping(
            GetListenKeys,
            name_style=NameStyle.CAMEL,
            omit_default=True,
        ),
        name_mapping(
            ExtendListenKey,
            name_style=NameStyle.CAMEL,
            omit_default=True,
        ),
        name_mapping(
            DeleteListenKey,
            name_style=NameStyle.CAMEL,
            omit_default=True,
        ),
    ]
    + type_recipes,
)

__all__ = ["_retort"]



from aiomexc.methods import (
    GetTickerPrice,
    GetAccountInformation,
    QueryOrder,
    MexcMethod,
    CreateListenKey,
    GetListenKeys,
    ExtendListenKey,
    DeleteListenKey,
    GetOpenOrders,
)
from aiomexc.types import (
    TickerPrice,
    AccountInformation,
    Order,
    MexcType,
    ListenKey,
    ListenKeys,
)

from .session.base import BaseSession, Credentials
from .session.aiohttp import AiohttpSession


class MexcClient:
    def __init__(
        self, credentials: Credentials | None = None, session: BaseSession | None = None
    ):
        if session is None:
            session = AiohttpSession()

        self.session = session
        self.credentials = credentials

    async def __call__(
        self, method: MexcMethod[MexcType], credentials: Credentials | None = None
    ) -> MexcType:
        if method.__requires_auth__:
            # Use request-specific credentials if provided, otherwise fall back to session credentials
            credentials = credentials or self.credentials
            if credentials is None:
                raise ValueError(
                    f"Credentials are required for {method.__api_method__!r} method"
                )

            return await self.session.make_signed_request(method, credentials)
        else:
            return await self.session.make_request(method)

    async def get_ticker_price(self, symbol: str) -> TickerPrice:
        return await self(GetTickerPrice(symbol=symbol))

    async def get_account_information(
        self, credentials: Credentials | None = None
    ) -> AccountInformation:
        return await self(GetAccountInformation(), credentials=credentials)

    async def query_order(
        self,
        symbol: str,
        order_id: str | None = None,
        orig_client_order_id: str | None = None,
        credentials: Credentials | None = None,
    ) -> Order:
        return await self(
            QueryOrder(
                symbol=symbol,
                order_id=order_id,
                orig_client_order_id=orig_client_order_id,
            ),
            credentials=credentials,
        )

    async def create_listen_key(
        self, credentials: Credentials | None = None
    ) -> ListenKey:
        return await self(CreateListenKey(), credentials=credentials)

    async def get_listen_keys(
        self, credentials: Credentials | None = None
    ) -> ListenKeys:
        return await self(GetListenKeys(), credentials=credentials)

    async def extend_listen_key(
        self, listen_key: str, credentials: Credentials | None = None
    ) -> ListenKey:
        return await self(
            ExtendListenKey(listen_key=listen_key), credentials=credentials
        )

    async def delete_listen_key(
        self, listen_key: str, credentials: Credentials | None = None
    ) -> ListenKey:
        return await self(
            DeleteListenKey(listen_key=listen_key), credentials=credentials
        )

    async def get_open_orders(
        self, symbol: str, credentials: Credentials | None = None
    ) -> list[Order]:
        return await self(GetOpenOrders(symbol=symbol), credentials=credentials)
